from contextlib import redirect_stdout
from datetime import datetime
from inspect import Parameter, signature
from logging import Logger
from mage_ai.data_cleaner.shared.utils import (
    is_dataframe,
    is_geo_dataframe,
)
from mage_ai.data_preparation.logger_manager import StreamToLogger
from mage_ai.data_preparation.models.block.sql import execute_sql_code
from mage_ai.data_preparation.models.constants import (
    BlockLanguage,
    BlockStatus,
    BlockType,
    ExecutorType,
    CUSTOM_EXECUTION_BLOCK_TYPES,
    DATAFRAME_ANALYSIS_MAX_COLUMNS,
    DATAFRAME_ANALYSIS_MAX_ROWS,
    DATAFRAME_SAMPLE_COUNT_PREVIEW,
    NON_PIPELINE_EXECUTABLE_BLOCK_TYPES,
)
from mage_ai.data_preparation.models.file import File
from mage_ai.data_preparation.models.variable import VariableType
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.data_preparation.templates.template import load_template
from mage_ai.server.kernel_output_parser import DataType
from mage_ai.shared.constants import ENV_DEV
from mage_ai.shared.logger import BlockFunctionExec
from mage_ai.shared.parsers import encode_complex
from mage_ai.shared.utils import clean_name
from queue import Queue
from typing import Callable, Dict, List, Set
import asyncio
import functools
import json
import os
import pandas as pd
import simplejson
import sys
import time
import traceback


async def run_blocks(
    root_blocks: List['Block'],
    analyze_outputs: bool = True,
    build_block_output_stdout: Callable[..., object] = None,
    global_vars=None,
    parallel: bool = True,
    run_sensors: bool = True,
    run_tests: bool = True,
    selected_blocks: Set[str] = None,
    update_status: bool = True,
) -> None:
    tries_by_block_uuid = {}
    tasks = dict()
    blocks = Queue()

    def create_block_task(block: 'Block'):
        async def execute_and_run_tests():
            with BlockFunctionExec(
                block.uuid,
                f'Executing {block.type} block...',
                build_block_output_stdout=build_block_output_stdout,
            ):
                await block.execute(
                    analyze_outputs=analyze_outputs,
                    build_block_output_stdout=build_block_output_stdout,
                    global_vars=global_vars,
                    run_all_blocks=True,
                    run_sensors=run_sensors,
                    update_status=update_status,
                    parallel=parallel,
                )
                if run_tests:
                    block.run_tests(
                        build_block_output_stdout=build_block_output_stdout,
                        update_tests=False,
                    )

        return asyncio.create_task(execute_and_run_tests())

    for block in root_blocks:
        blocks.put(block)
        tasks[block.uuid] = None

    while not blocks.empty():
        block = blocks.get()

        if block.type in NON_PIPELINE_EXECUTABLE_BLOCK_TYPES or \
                not run_sensors and block.type == BlockType.SENSOR:
            continue

        if tries_by_block_uuid.get(block.uuid, None) is None:
            tries_by_block_uuid[block.uuid] = 0

        tries_by_block_uuid[block.uuid] += 1
        tries = tries_by_block_uuid[block.uuid]
        if tries >= 1000:
            raise Exception(f'Block {block.uuid} has tried to execute {tries} times; exiting.')

        skip = False
        for upstream_block in block.upstream_blocks:
            if tasks.get(upstream_block.uuid) is None:
                blocks.put(block)
                skip = True
                break
        if skip:
            continue
        upstream_tasks = [tasks[u.uuid] for u in block.upstream_blocks]
        await asyncio.gather(*upstream_tasks)
        block_task = create_block_task(block)
        tasks[block.uuid] = block_task
        for downstream_block in block.downstream_blocks:
            if downstream_block.uuid not in tasks and (
                selected_blocks is None or upstream_block.uuid in selected_blocks
            ):

                tasks[downstream_block.uuid] = None
                blocks.put(downstream_block)
    remaining_tasks = filter(lambda task: task is not None, tasks.values())
    await asyncio.gather(*remaining_tasks)


def run_blocks_sync(
    root_blocks: List['Block'],
    analyze_outputs: bool = True,
    build_block_output_stdout: Callable[..., object] = None,
    global_vars: Dict = None,
    run_sensors: bool = True,
    run_tests: bool = True,
    selected_blocks: Set[str] = None,
) -> None:
    tries_by_block_uuid = {}
    tasks = dict()
    blocks = Queue()

    for block in root_blocks:
        blocks.put(block)
        tasks[block.uuid] = False

    while not blocks.empty():
        block = blocks.get()

        if block.type in NON_PIPELINE_EXECUTABLE_BLOCK_TYPES:
            continue

        if tries_by_block_uuid.get(block.uuid, None) is None:
            tries_by_block_uuid[block.uuid] = 0

        tries_by_block_uuid[block.uuid] += 1
        tries = tries_by_block_uuid[block.uuid]
        if tries >= 1000:
            raise Exception(f'Block {block.uuid} has tried to execute {tries} times; exiting.')

        skip = False
        for upstream_block in block.upstream_blocks:
            upstream_task_status = tasks.get(upstream_block.uuid)
            if upstream_task_status is None or not upstream_task_status:
                blocks.put(block)
                skip = True
                break
        if skip:
            continue
        with BlockFunctionExec(
            block.uuid,
            f'Executing {block.type} block...',
            build_block_output_stdout=build_block_output_stdout,
        ):
            block.execute_sync(
                analyze_outputs=analyze_outputs,
                build_block_output_stdout=build_block_output_stdout,
                global_vars=global_vars,
                run_all_blocks=True,
                test_execution=not run_sensors,
            )
            if run_tests:
                block.run_tests(
                    build_block_output_stdout=build_block_output_stdout,
                    update_tests=False,
                )
        tasks[block.uuid] = True
        for downstream_block in block.downstream_blocks:
            if downstream_block.uuid not in tasks and (
                selected_blocks is None or downstream_block.uuid in selected_blocks
            ):
                tasks[downstream_block.uuid] = None
                blocks.put(downstream_block)


class Block:
    def __init__(
        self,
        name: str,
        uuid: str,
        block_type: BlockType,
        content: str = None,
        executor_config: Dict = None,
        executor_type: ExecutorType = ExecutorType.LOCAL_PYTHON,
        status: BlockStatus = BlockStatus.NOT_EXECUTED,
        pipeline=None,
        language: BlockLanguage = BlockLanguage.PYTHON,
        configuration: Dict = dict(),
    ):
        self.name = name or uuid
        self.uuid = uuid
        self.type = block_type
        self._content = content
        self.executor_config = executor_config
        self.executor_type = executor_type
        self.status = status
        self.pipeline = pipeline
        self.language = language
        self.configuration = configuration

        self._outputs = None
        self._outputs_loaded = False
        self.upstream_blocks = []
        self.downstream_blocks = []
        self.test_functions = []
        self.global_vars = {}

    @property
    def content(self):
        if self._content is None:
            self._content = self.file.content()
        return self._content

    @property
    def executable(self):
        return self.type not in NON_PIPELINE_EXECUTABLE_BLOCK_TYPES

    @property
    def input_variables(self):
        return {b.uuid: b.output_variables.keys() for b in self.upstream_blocks}

    @property
    def output_variables(self):
        """
        Return output variables in dictionary.
        The key is the variable name, and the value is variable data type.
        """
        return dict()

    @property
    def outputs(self):
        if not self._outputs_loaded:
            if self._outputs is None or len(self._outputs) == 0:
                self._outputs = self.get_outputs()
        return self._outputs

    @property
    def upstream_block_uuids(self):
        return [b.uuid for b in self.upstream_blocks]

    @property
    def downstream_block_uuids(self):
        return [b.uuid for b in self.downstream_blocks]

    @property
    def file_path(self):
        repo_path = self.pipeline.repo_path if self.pipeline is not None else get_repo_path()
        file_extension = 'sql' if BlockLanguage.SQL == self.language else 'py'

        return os.path.join(
            repo_path or os.getcwd(),
            f'{self.type}s/{self.uuid}.{file_extension}',
        )

    @property
    def file(self):
        return File.from_path(self.file_path)

    @property
    def table_name(self):
        table_name = f'{self.pipeline.uuid}_{self.uuid}_{self.pipeline.version_name}'

        env = (self.global_vars or dict()).get('env')
        if env == ENV_DEV:
            table_name = f'dev_{table_name}'

        return table_name

    @classmethod
    def block_class_from_type(self, block_type: str) -> str:
        return BLOCK_TYPE_TO_CLASS.get(block_type)

    @classmethod
    def after_create(self, block, **kwargs):
        widget = kwargs.get('widget')
        pipeline = kwargs.get('pipeline')
        if pipeline is not None:
            priority = kwargs.get('priority')
            upstream_block_uuids = kwargs.get('upstream_block_uuids')
            pipeline.add_block(block, upstream_block_uuids, priority=priority, widget=widget)

    @classmethod
    def create(
        self,
        name,
        block_type,
        repo_path,
        configuration=None,
        language=None,
        pipeline=None,
        priority=None,
        upstream_block_uuids=None,
        config=None,
        widget=False,
    ):
        """
        1. Create a new folder for block_type if not exist
        2. Create a new python file with code template
        """
        if upstream_block_uuids is None:
            upstream_block_uuids = []
        if config is None:
            config = {}

        uuid = clean_name(name)
        block_dir_path = os.path.join(repo_path, f'{block_type}s')
        if not os.path.exists(block_dir_path):
            os.mkdir(block_dir_path)
            with open(os.path.join(block_dir_path, '__init__.py'), 'w'):
                pass

        file_extension = 'sql' if BlockLanguage.SQL == language else 'py'
        file_path = os.path.join(block_dir_path, f'{uuid}.{file_extension}')
        if os.path.exists(file_path):
            if pipeline is not None and pipeline.has_block(uuid):
                raise Exception(f'Block {uuid} already exists. Please use a different name.')
        else:
            load_template(
                block_type,
                config,
                file_path,
                language=language,
                pipeline_type=pipeline.type if pipeline is not None else None,
            )

        block = self.block_class_from_type(block_type)(
            name,
            uuid,
            block_type,
            configuration=configuration,
            language=language,
            pipeline=pipeline,
        )
        self.after_create(
            block,
            config=config,
            pipeline=pipeline,
            priority=priority,
            upstream_block_uuids=upstream_block_uuids,
            widget=widget,
        )
        return block

    @classmethod
    def get_all_blocks(self, repo_path):
        block_uuids = dict()
        for t in BlockType:
            block_dir = os.path.join(repo_path, f'{t.value}s')
            if not os.path.exists(block_dir):
                continue
            block_uuids[t.value] = []
            for f in os.listdir(block_dir):
                if (f.endswith('.py') or f.endswith('.sql')) and f != '__init__.py':
                    block_uuids[t.value].append(f.split('.')[0])
        return block_uuids

    @classmethod
    def get_block(
        self,
        name,
        uuid,
        block_type,
        configuration=None,
        content=None,
        language=None,
        pipeline=None,
        status=BlockStatus.NOT_EXECUTED,
    ):
        block_class = self.block_class_from_type(block_type) or Block
        return block_class(
            name,
            uuid,
            block_type,
            configuration=configuration,
            content=content,
            language=language,
            pipeline=pipeline,
            status=status,
        )

    def all_upstream_blocks_completed(self, completed_block_uuids: Set[str]):
        return all(b.uuid in completed_block_uuids for b in self.upstream_blocks)

    def delete(self, widget=False, commit=True):
        """
        1. If pipeline is not None, delete the block from the pipeline but not delete the block
        file.
        2. If pipeline is None, check whether block is used in any pipelines. If block is being
        used, throw error. Otherwise, delete the block files.
        """
        from mage_ai.data_preparation.models.pipeline import Pipeline

        if self.pipeline is not None:
            self.pipeline.delete_block(self, widget=widget, commit=commit)
            # For block_type SCRATCHPAD, also delete the file if possible
            if self.type in NON_PIPELINE_EXECUTABLE_BLOCK_TYPES:
                pipelines = Pipeline.get_pipelines_by_block(self, widget=widget)
                pipelines = [
                    pipeline for pipeline in pipelines if self.pipeline.uuid != pipeline.uuid
                ]
                if len(pipelines) == 0:
                    os.remove(self.file_path)
            return
        # If pipeline is not specified, delete the block from all pipelines and delete the file.
        pipelines = Pipeline.get_pipelines_by_block(self, widget=widget)
        for p in pipelines:
            if not p.block_deletable(self):
                raise Exception(
                    f'Block {self.uuid} has downstream dependencies in pipeline {p.uuid}. '
                    'Please remove the dependencies before deleting the block.'
                )
        for p in pipelines:
            p.delete_block(p.get_block(self.uuid, widget=widget), widget=widget, commit=commit)
        os.remove(self.file_path)

    def execute_sync(
        self,
        analyze_outputs: bool = True,
        build_block_output_stdout: Callable[..., object] = None,
        custom_code: str = None,
        execution_partition: str = None,
        global_vars: Dict = None,
        logger: Logger = None,
        run_all_blocks: bool = False,
        test_execution: bool = False,
        update_status: bool = True,
    ):
        try:
            if not run_all_blocks:
                not_executed_upstream_blocks = list(
                    filter(lambda b: b.status == BlockStatus.NOT_EXECUTED, self.upstream_blocks)
                )
                if len(not_executed_upstream_blocks) > 0:
                    raise Exception(
                        f"Block {self.uuid}'s upstream blocks have not been executed yet. "
                        f'Please run upstream blocks {list(map(lambda b: b.uuid, not_executed_upstream_blocks))} '
                        'before running the current block.'
                    )
            output = self.execute_block(
                build_block_output_stdout=build_block_output_stdout,
                custom_code=custom_code,
                execution_partition=execution_partition,
                global_vars=global_vars,
                logger=logger,
                test_execution=test_execution,
            )
            block_output = output['output']
            if BlockType.CHART == self.type:
                variable_mapping = block_output
                output = dict(
                    output=simplejson.dumps(
                        block_output,
                        default=encode_complex,
                        ignore_nan=True,
                    )
                )
            elif BlockType.SENSOR == self.type:
                variable_mapping = dict()
            else:
                self.__verify_outputs(block_output)
                variable_keys = list(self.output_variables.keys())
                extra_output_count = len(block_output) - len(variable_keys)
                variable_keys += [f'output_{idx}' for idx in range(extra_output_count)]
                variable_mapping = dict(zip(variable_keys, block_output))

            self.store_variables(
                variable_mapping,
                execution_partition=execution_partition,
                override_outputs=True,
                spark=(global_vars or dict()).get('spark'),
            )
            # Reset outputs cache
            self._outputs = None

            if update_status:
                self.status = BlockStatus.EXECUTED

            if analyze_outputs and BlockType.CHART != self.type:
                self.analyze_outputs(variable_mapping)
        except Exception as err:
            if update_status:
                self.status = BlockStatus.FAILED
            if logger is not None:
                logger.exception(
                    f'Failed to execute block {self.uuid}',
                    block_type=self.type,
                    block_uuid=self.uuid,
                    error=err,
                )
            raise err
        finally:
            if update_status:
                self.__update_pipeline_block(widget=BlockType.CHART == self.type)

        return output

    async def execute(
        self,
        analyze_outputs: bool = True,
        build_block_output_stdout: Callable[..., object] = None,
        custom_code: str = None,
        global_vars=None,
        run_all_blocks: bool = False,
        run_sensors: bool = True,
        update_status: bool = True,
        parallel: bool = True,
    ) -> None:
        if parallel:
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(
                None,
                functools.partial(
                    self.execute_sync,
                    analyze_outputs=analyze_outputs,
                    build_block_output_stdout=build_block_output_stdout,
                    custom_code=custom_code,
                    global_vars=global_vars,
                    run_all_blocks=run_all_blocks,
                    test_execution=not run_sensors,
                    update_status=update_status,
                )
            )
        else:
            self.execute_sync(
                analyze_outputs=analyze_outputs,
                build_block_output_stdout=build_block_output_stdout,
                custom_code=custom_code,
                global_vars=global_vars,
                run_all_blocks=run_all_blocks,
                test_execution=not run_sensors,
                update_status=update_status,
            )

    def __validate_execution(self, decorated_functions, input_vars):
        if self.type not in CUSTOM_EXECUTION_BLOCK_TYPES:
            return None

        if BlockLanguage.SQL == self.language:
            return None

        if len(decorated_functions) == 0:
            raise Exception(
                f'Block {self.uuid} does not have any decorated functions. '
                f'Make sure that a function in the block is decorated with @{self.type}.'
            )
        else:
            block_function = decorated_functions[0]
            sig = signature(block_function)

            num_args = sum(
                arg.kind not in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD)
                for arg in sig.parameters.values()
            )
            num_inputs = len(input_vars)
            num_upstream = len(self.upstream_block_uuids)

            has_var_args = num_args != len(sig.parameters)

            if num_args > num_inputs:
                if num_upstream < num_args:
                    raise Exception(
                        f'Block {self.uuid} may be missing upstream dependencies. '
                        f'It expected to have {"at least " if has_var_args else ""}{num_args} arguments, '
                        f'but only received {num_inputs}. '
                        f'Confirm that the @{self.type} method declaration has the correct number of arguments.'
                    )
                else:
                    raise Exception(
                        f'Block {self.uuid} is missing input arguments. '
                        f'It expected to have {"at least " if has_var_args else ""}{num_args} arguments, '
                        f'but only received {num_inputs}. '
                        f'Double check the @{self.type} method declaration has the correct number of arguments '
                        f'and that the upstream blocks have been executed.'
                    )
            elif num_args < num_inputs and not has_var_args:
                if num_upstream > num_args:
                    raise Exception(
                        f'Block {self.uuid} may have too many upstream dependencies. '
                        f'It expected to have {num_args} arguments, but received {num_inputs}. '
                        f'Confirm that the @{self.type} method declaration has the correct number of arguments.'
                    )
                else:
                    raise Exception(
                        f'Block {self.uuid} has too many input arguments. '
                        f'It expected to have {num_args} arguments, but received {num_inputs}. '
                        f'Confirm that the @{self.type} method declaration has the correct number of arguments.'
                    )

            return block_function

    def execute_block(
        self,
        build_block_output_stdout: Callable[..., object] = None,
        custom_code: str = None,
        execution_partition: str = None,
        logger: Logger = None,
        global_vars: Dict = None,
        test_execution: bool = False,
    ) -> Dict:
        upstream_block_uuids = []
        input_vars = []
        if self.pipeline is not None:
            for upstream_block_uuid, variables in self.input_variables.items():
                upstream_block_uuids.append(upstream_block_uuid)
                input_vars += [
                    self.pipeline.variable_manager.get_variable(
                        self.pipeline.uuid,
                        upstream_block_uuid,
                        var,
                        partition=execution_partition,
                        variable_type=VariableType.DATAFRAME,
                        spark=(global_vars or dict()).get('spark'),
                    )
                    for var in variables
                ]
        outputs = []
        decorated_functions = []
        test_functions = []
        if logger is not None:
            stdout = StreamToLogger(logger)
        elif build_block_output_stdout:
            stdout = build_block_output_stdout(self.uuid)
        else:
            stdout = sys.stdout
        results = {}
        outputs_from_input_vars = {}

        for idx, input_var in enumerate(input_vars):
            upstream_block_uuid = upstream_block_uuids[idx]
            outputs_from_input_vars[upstream_block_uuid] = input_var
            outputs_from_input_vars[f'df_{idx + 1}'] = input_var

        with redirect_stdout(stdout):
            results = {
                self.type: self.__block_decorator(decorated_functions),
                'test': self.__block_decorator(test_functions),
            }
            results.update(outputs_from_input_vars)

            outputs = []

            if BlockLanguage.SQL == self.language and BlockType.CHART != self.type:
                outputs = execute_sql_code(
                    self,
                    custom_code or self.content,
                    execution_partition=execution_partition,
                    global_vars=global_vars,
                )
            elif custom_code is not None:
                if BlockType.CHART != self.type or (not self.group_by_columns or not self.metrics):
                    exec(custom_code, results)
            elif self.content is not None:
                exec(self.content, results)
            elif os.path.exists(self.file_path):
                with open(self.file_path) as file:
                    exec(file.read(), results)

            self.test_functions = test_functions

            if BlockType.CHART == self.type:
                variables = self.get_variables_from_code_execution(results)
                outputs = self.post_process_variables(
                    variables,
                    code=custom_code,
                    results=results,
                    upstream_block_uuids=upstream_block_uuids,
                )
            else:
                block_function = self.__validate_execution(decorated_functions, input_vars)
                if block_function is not None:
                    outputs = self.execute_block_function(block_function, input_vars, global_vars, test_execution)

                if outputs is None:
                    outputs = []
                if type(outputs) is not list:
                    outputs = [outputs]

        output_message = dict(output=outputs)

        return output_message

    def execute_block_function(
        self,
        block_function: Callable,
        input_vars: List,
        global_vars: Dict = None,
        test_execution: bool = False,
    ) -> Dict:
        sig = signature(block_function)
        has_kwargs = any([p.kind == p.VAR_KEYWORD for p in sig.parameters.values()])
        if has_kwargs and global_vars is not None and len(global_vars) != 0:
            output = block_function(*input_vars, **global_vars)
        else:
            output = block_function(*input_vars)
        return output

    def exists(self):
        return os.path.exists(self.file_path)

    def get_analyses(self):
        if self.status == BlockStatus.NOT_EXECUTED:
            return []
        if len(self.output_variables) == 0:
            return []
        analyses = []
        for v, vtype in self.output_variables.items():
            if vtype is not pd.DataFrame:
                continue
            data = self.pipeline.variable_manager.get_variable(
                self.pipeline.uuid,
                self.uuid,
                v,
                variable_type=VariableType.DATAFRAME_ANALYSIS,
            )
            data['variable_uuid'] = v
            analyses.append(data)
        return analyses

    def get_outputs(
        self,
        execution_partition: str = None,
        include_print_outputs: bool = True,
        sample_count: int = DATAFRAME_SAMPLE_COUNT_PREVIEW,
    ) -> List[Dict]:
        if self.pipeline is None:
            return
        if self.type != BlockType.SCRATCHPAD and BlockType.CHART != self.type:
            if self.status == BlockStatus.NOT_EXECUTED:
                return []

        data_products = []
        outputs = []
        variable_manager = self.pipeline.variable_manager

        all_variables = variable_manager.get_variables_by_block(
            self.pipeline.uuid,
            self.uuid,
            partition=execution_partition,
        )

        if not include_print_outputs:
            all_variables = [v for v in all_variables if v in self.output_variables.keys()
                             or v.startswith('output')]

        for v in all_variables:
            data = variable_manager.get_variable(
                self.pipeline.uuid,
                self.uuid,
                v,
                partition=execution_partition,
                sample=True,
                sample_count=sample_count,
            )
            if type(data) is pd.DataFrame:
                try:
                    analysis = variable_manager.get_variable(
                        self.pipeline.uuid,
                        self.uuid,
                        v,
                        partition=execution_partition,
                        variable_type=VariableType.DATAFRAME_ANALYSIS,
                    )
                except Exception:
                    analysis = None
                if analysis is not None:
                    stats = analysis.get('statistics', {})
                    column_types = analysis.get('metadata', {}).get('column_types', {})
                    row_count = stats.get('original_row_count', stats.get('count'))
                    column_count = len(column_types)
                else:
                    row_count, column_count = data.shape

                columns_to_display = data.columns.tolist()[:DATAFRAME_ANALYSIS_MAX_COLUMNS]
                data = dict(
                    sample_data=dict(
                        columns=columns_to_display,
                        rows=json.loads(data[columns_to_display].to_json(orient='split'))['data']
                    ),
                    shape=[row_count, column_count],
                    type=DataType.TABLE,
                    variable_uuid=v,
                )
                data_products.append(data)
                continue
            elif is_geo_dataframe(data):
                data = dict(
                    text_data=f''' Use the following code in a scratchpad to get the output of the block:

from mage_ai.data_preparation.variable_manager import get_variable
df = get_variable('{self.pipeline.uuid}', '{self.uuid}', 'df')
''',
                    type=DataType.TEXT,
                    variable_uuid=v,
                )
            elif type(data) is str:
                data = dict(
                    text_data=data,
                    type=DataType.TEXT,
                    variable_uuid=v,
                )
            elif type(data) is dict or type(data) is list:
                data = dict(
                    text_data=simplejson.dumps(
                        data,
                        default=datetime.isoformat,
                        ignore_nan=True,
                    ),
                    type=DataType.TEXT,
                    variable_uuid=v,
                )
            outputs.append(data)
        return outputs + data_products

    def save_outputs(self, outputs, override=False):
        variable_mapping = dict()
        for o in outputs:
            if o is None:
                continue
            if all(k in o for k in ['variable_uuid', 'text_data']) and \
                    o['variable_uuid'] != 'df':
                variable_mapping[o['variable_uuid']] = o['text_data']

        self._outputs = outputs
        self._outputs_loaded = True
        self.store_variables(variable_mapping, override=override)

    def to_dict(self, include_content=False, include_outputs=False, sample_count=None):
        language = self.language
        if language and type(self.language) is not str:
            language = self.language.value

        def __format_enum(v):
            return v.value if type(v) is not str else v

        data = dict(
            all_upstream_blocks_executed=all(
                block.status == BlockStatus.EXECUTED for block in self.get_all_upstream_blocks()
            ),
            configuration=self.configuration or {},
            downstream_blocks=self.downstream_block_uuids,
            executor_config=self.executor_config,
            executor_type=__format_enum(self.executor_type),
            name=self.name,
            language=language,
            status=__format_enum(self.status),
            type=__format_enum(self.type),
            upstream_blocks=self.upstream_block_uuids,
            uuid=self.uuid,
        )
        if include_content:
            data['content'] = self.content
        if include_outputs:
            data['outputs'] = self.outputs
        return data

    def update(self, data):
        if 'name' in data and data['name'] != self.name:
            self.__update_name(data['name'])
        if (
            'type' in data
            and self.type == BlockType.SCRATCHPAD
            and data['type'] != BlockType.SCRATCHPAD
        ):
            self.__update_type(data['type'])
        if 'upstream_blocks' in data and set(data['upstream_blocks']) != set(
            self.upstream_block_uuids
        ):
            self.__update_upstream_blocks(data['upstream_blocks'])
        if 'executor_type' in data and data['executor_type'] != self.executor_type:
            self.executor_type = data['executor_type']
            self.__update_pipeline_block()
        return self

    def update_content(self, content, widget=False):
        if content != self.content:
            self.status = BlockStatus.UPDATED
            self._content = content
            self.file.update_content(content)
            self.__update_pipeline_block(widget=widget)
        return self

    def update_status(self, status: BlockStatus):
        self.status = status
        self.__update_pipeline_block(widget=BlockType.CHART == self.type)

    def get_all_upstream_blocks(self) -> List['Block']:
        queue = Queue()
        visited = set()
        queue.put(self)
        while not queue.empty():
            current_block = queue.get()
            for block in current_block.upstream_blocks:
                if block.uuid not in visited:
                    queue.put(block)
                    visited.add(block)
        return visited

    def run_upstream_blocks(self) -> None:
        def process_upstream_block(
            block: 'Block',
            root_blocks: List['Block'],
        ) -> List[str]:
            if len(block.upstream_blocks) == 0:
                root_blocks.append(block)
            return block.uuid

        upstream_blocks = self.get_all_upstream_blocks()
        root_blocks = []
        upstream_block_uuids = list(
            map(lambda x: process_upstream_block(x, root_blocks), upstream_blocks)
        )

        run_blocks_sync(root_blocks, selected_blocks=upstream_block_uuids)

    def run_tests(
        self,
        build_block_output_stdout: Callable[..., object] = None,
        custom_code: str = None,
        logger: Logger = None,
        update_tests: bool = True,
    ) -> None:
        test_functions = []
        if update_tests:
            results = {
                'test': self.__block_decorator(test_functions),
            }
            if custom_code is not None:
                exec(custom_code, results)
            elif os.path.exists(self.file_path):
                with open(self.file_path) as file:
                    exec(file.read(), results)
        else:
            test_functions = self.test_functions

        variable_manager = self.pipeline.variable_manager
        outputs = [
            variable_manager.get_variable(
                self.pipeline.uuid,
                self.uuid,
                variable,
            )
            for variable in self.output_variables.keys()
        ]

        if logger is not None:
            stdout = StreamToLogger(logger)
        elif build_block_output_stdout:
            stdout = build_block_output_stdout(self.uuid)
        else:
            stdout = sys.stdout

        with redirect_stdout(stdout):
            tests_passed = 0
            for func in test_functions:
                try:
                    func(*outputs)
                    tests_passed += 1
                except AssertionError:
                    print('==============================================================')
                    print(f'FAIL: {func.__name__} (block: {self.uuid})')
                    print('--------------------------------------------------------------')
                    print(traceback.format_exc())
            print('--------------------------------------------------------------')
            print(f'{tests_passed}/{len(test_functions)} tests passed.')
        if tests_passed != len(test_functions):
            raise Exception(f'Failed to pass tests for block {self.uuid}')

    def analyze_outputs(self, variable_mapping):
        from mage_ai.data_cleaner.data_cleaner import clean as clean_data

        if self.pipeline is None:
            return
        for uuid, data in variable_mapping.items():
            vtype = self.output_variables.get(uuid)
            if vtype is pd.DataFrame:
                if data.shape[1] > DATAFRAME_ANALYSIS_MAX_COLUMNS:
                    continue
                if data.shape[0] > DATAFRAME_ANALYSIS_MAX_ROWS:
                    data_for_analysis = data.sample(DATAFRAME_ANALYSIS_MAX_ROWS).reset_index(
                        drop=True,
                    )
                else:
                    data_for_analysis = data.reset_index(drop=True)
                try:
                    analysis = clean_data(
                        data_for_analysis,
                        df_original=data,
                        transform=False,
                        verbose=False,
                    )
                    self.pipeline.variable_manager.add_variable(
                        self.pipeline.uuid,
                        self.uuid,
                        uuid,
                        dict(
                            metadata=dict(column_types=analysis['column_types']),
                            statistics=analysis['statistics'],
                            insights=analysis['insights'],
                            suggestions=analysis['suggestions'],
                        ),
                        variable_type=VariableType.DATAFRAME_ANALYSIS,
                    )
                except Exception:
                    pass
                    # TODO: we use to silently fail, but it looks bad when using BigQuery
                    # print('\nFailed to analyze dataframe:')
                    # print(traceback.format_exc())

    def set_global_vars(self, global_vars):
        self.global_vars = global_vars
        for upstream_block in self.upstream_blocks:
            upstream_block.global_vars = global_vars

    def store_variables(
        self,
        variable_mapping: Dict,
        execution_partition: str = None,
        override: bool = False,
        override_outputs: bool = False,
        spark=None,
    ):
        if self.pipeline is None:
            return
        all_variables = self.pipeline.variable_manager.get_variables_by_block(
            self.pipeline.uuid,
            self.uuid,
            partition=execution_partition,
        )
        removed_variables = []
        if override:
            # Not remove dataframe variables
            removed_variables = [v for v in all_variables
                                 if v not in variable_mapping.keys()
                                 and v != 'df'
                                 and not v.startswith('output')]
        elif override_outputs:
            removed_variables = [v for v in all_variables
                                 if v not in variable_mapping.keys() and v.startswith('output')]
        for uuid, data in variable_mapping.items():
            if spark is not None and type(data) is pd.DataFrame:
                data = spark.createDataFrame(data)
            self.pipeline.variable_manager.add_variable(
                self.pipeline.uuid,
                self.uuid,
                uuid,
                data,
                partition=execution_partition,
            )

        for uuid in removed_variables:
            self.pipeline.variable_manager.delete_variable(
                self.pipeline.uuid,
                self.uuid,
                uuid,
            )

    # TODO: Update all pipelines that use this block
    def __update_name(self, name):
        """
        1. Rename block file
        2. Update the folder of variable
        3. Update upstream and downstream relationships
        """
        old_uuid = self.uuid
        old_file_path = self.file_path
        new_uuid = clean_name(name)
        self.name = name
        self.uuid = new_uuid
        new_file_path = self.file_path
        if self.pipeline is not None:
            if self.pipeline.has_block(new_uuid):
                raise Exception(
                    f'Block {new_uuid} already exists in pipeline. Please use a different name.'
                )
        if os.path.exists(new_file_path):
            raise Exception(f'Block {new_uuid} already exists. Please use a different name.')
        os.rename(old_file_path, new_file_path)
        if self.pipeline is not None:
            self.pipeline.update_block_uuid(self, old_uuid)

    def __update_pipeline_block(self, widget=False):
        if self.pipeline is None:
            return
        self.pipeline.update_block(self, widget=widget)

    def __update_type(self, block_type):
        """
        1. Move block file to another folder
        2. Update the block type in pipeline metadata.yaml
        3. Update the code in block file
        """
        old_file_path = self.file_path
        self.type = block_type
        new_file_path = self.file_path
        if os.path.exists(new_file_path):
            raise Exception(
                f'Block {self.type}/{self.uuid} already exists.'
                ' Please rename it before changing the type.'
            )
        os.rename(old_file_path, new_file_path)
        if self.pipeline is not None:
            self.pipeline.update_block(self)
        with open(new_file_path) as f:
            existing_code = f.read()
        load_template(
            block_type,
            dict(existing_code='    ' + existing_code.replace('\n', '\n    ')),
            new_file_path,
            language=self.language,
        )

    def __update_upstream_blocks(self, upstream_blocks):
        if self.pipeline is None:
            return
        self.pipeline.update_block(
            self,
            upstream_block_uuids=upstream_blocks,
            widget=BlockType.CHART == self.type,
        )

    def __verify_outputs(self, outputs):
        if len(outputs) < len(self.output_variables):
            raise Exception(
                f'Validation error for block {self.uuid}: '
                f'the number of output variables does not match the block type: {self.type} ',
            )
        variable_names = list(self.output_variables.keys())
        variable_dtypes = list(self.output_variables.values())
        for idx, output in enumerate(outputs):
            if idx >= len(variable_dtypes):
                break
            actual_dtype = type(output)
            expected_dtype = variable_dtypes[idx]
            if (expected_dtype != pd.DataFrame and actual_dtype is not expected_dtype) or (
                expected_dtype == pd.DataFrame and not is_dataframe(output)
            ):
                raise Exception(
                    f'Validation error for block {self.uuid}: '
                    f'the variable {variable_names[idx]} should be {expected_dtype} type, '
                    f'but {actual_dtype} type is returned',
                )

    def __block_decorator(self, decorated_functions):
        def custom_code(function):
            decorated_functions.append(function)
            return function

        return custom_code


class DataLoaderBlock(Block):
    @property
    def output_variables(self):
        return dict(df=pd.DataFrame)


class DataExporterBlock(Block):
    @property
    def output_variables(self):
        return dict()


class TransformerBlock(Block):
    @property
    def output_variables(self):
        return dict(df=pd.DataFrame)


class SensorBlock(Block):
    @property
    def output_variables(self):
        return dict()

    def execute_block_function(
        self,
        block_function: Callable,
        input_vars: List,
        global_vars: Dict = None,
        test_execution: bool = False,
    ) -> List:
        if test_execution:
            return super().execute_block_function(
                block_function,
                input_vars,
                global_vars=global_vars,
                test_execution=test_execution,
            )
        else:
            sig = signature(block_function)
            has_kwargs = any([p.kind == p.VAR_KEYWORD for p in sig.parameters.values()])
            use_global_vars = has_kwargs and global_vars is not None and len(global_vars) != 0
            while True:
                condition = block_function(**global_vars) if use_global_vars else block_function()
                if condition:
                    break
                print('Sensor sleeping for 1 minute...')
                time.sleep(60)
            return []


BLOCK_TYPE_TO_CLASS = {
    BlockType.DATA_EXPORTER: DataExporterBlock,
    BlockType.DATA_LOADER: DataLoaderBlock,
    BlockType.SCRATCHPAD: Block,
    BlockType.TRANSFORMER: TransformerBlock,
    BlockType.SENSOR: SensorBlock,
}



from typing import Dict, List
import json


def parse_runtime_variables(variables: List[str]) -> Dict[str, any]:
    vars_parsed = dict()

    if variables is None:
        return vars_parsed
    for i in range(0, len(variables), 2):
        key = variables[i]
        try:
            value = variables[i+1]
        except IndexError:
            value = None
        vars_parsed[key] = get_value(value)
    return vars_parsed


def get_value(value: str) -> any:
    if value is not None:
        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError:
            pass
    return value



from mage_ai.data_cleaner.shared.utils import wrap_column_name
from mage_ai.tests.base_test import TestCase


class SharedTests(TestCase):
    def test_wrap_column_name(self):
        column_names = [
            'good_name',
            'bad name',
            'name_with_symbols!%^&*()-+_=',
            'crazy _)_+(+_a aslfiewawlhi 3452',
            'AnotherGoodName',
            'NowNoGood(at_least_filtering_wise)',
        ]
        expected_names = [
            'good_name',
            '"bad name"',
            '"name_with_symbols!%^&*()-+_="',
            '"crazy _)_+(+_a aslfiewawlhi 3452"',
            'AnotherGoodName',
            '"NowNoGood(at_least_filtering_wise)"',
        ]
        for name, expected_name in zip(column_names, expected_names):
            self.assertEqual(wrap_column_name(name), expected_name)



from mage_ai.data_cleaner.cleaning_rules.base import BaseRule
from mage_ai.data_cleaner.transformer_actions.constants import (
    ActionType,
    Axis,
)
from mage_ai.data_cleaner.transformer_actions.utils import build_action_variables


class FixSyntaxErrors(BaseRule):
    def evaluate(self):
        columns_with_syntax_errors = []
        for column in self.df_columns:
            if self.statistics[f'{column}/invalid_value_rate'] > 0:
                columns_with_syntax_errors.append(column)

        suggestions = []
        if len(columns_with_syntax_errors):
            suggestions.append(
                self._build_transformer_action_suggestion(
                    title='Fix syntax errors',
                    message='Fix syntactical errors to reduce the amount of noise in the data.',
                    action_type=ActionType.FIX_SYNTAX_ERRORS,
                    action_arguments=columns_with_syntax_errors,
                    action_variables=build_action_variables(
                        self.df, self.column_types, columns_with_syntax_errors
                    ),
                    axis=Axis.COLUMN,
                )
            )
        return suggestions



{% extends "transformers/transformer_actions/action.jinja" %}
{% block action %}
    """
    Execute Transformer Action: ActionType.FIX_SYNTAX_ERRORS

    This marks any improperly formatted values in each column specified
    as invalid.

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/actions/transformer_actions/README.md#fix-syntax-errors
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.FIX_SYNTAX_ERRORS,
        arguments=df.columns,  # Specify columns to fix syntax errors for.
        axis=Axis.COLUMN,
    )
{% endblock %}



from dataclasses import dataclass
from mage_ai.services.email.config import EmailConfig
from mage_ai.services.slack.config import SlackConfig
from mage_ai.shared.config import BaseConfig
from typing import Dict
import traceback


@dataclass
class NotificationConfig(BaseConfig):
    email_config: EmailConfig = None
    slack_config: SlackConfig = None

    @classmethod
    def load(self, config_path: str = None, config: Dict = None):
        notification_config = super().load(config_path=config_path, config=config)
        if notification_config.slack_config is not None and \
                type(notification_config.slack_config) is dict:
            try:
                notification_config.slack_config = SlackConfig.load(
                    config=notification_config.slack_config,
                )
            except Exception:
                traceback.print_exc()
                notification_config.slack_config = None
        if notification_config.email_config is not None and \
                type(notification_config.email_config) is dict:
            try:
                notification_config.email_config = EmailConfig.load(
                    config=notification_config.email_config,
                )
            except Exception:
                traceback.print_exc()
                notification_config.email_config = None
        return notification_config



from mage_ai.data_cleaner.transformer_actions.udf.base import BaseUDF
import numpy as np
import pandas as pd


class DateTrunc(BaseUDF):
    def execute(self):
        date_part = self.options['date_part']
        date_column = self.arguments[0]
        df_copy = self.df.copy()
        df_copy[date_column] = pd.to_datetime(df_copy[date_column])
        if date_part == 'week':
            return (df_copy[date_column] - df_copy[date_column].dt.weekday * np.timedelta64(1, 'D')).\
                    dt.strftime('%Y-%m-%d')

        raise Exception(f'Date part {date_part} is not supported.')



"""Static definition of block decorators.
These decorators definitions are added to avoid python compilation errors.
The actual definitions of the decorators are generated at runtime dynamically in
execute_block method of mage_ai/data_preparation/models/block.py file.
"""


def data_exporter(function):
    return function


def data_loader(function):
    return function


def sensor(function):
    return function


def transformer(function):
    return function


def test(function):
    return function



{% extends "transformers/transformer_actions/action.jinja" %}
{% block imports %}
from mage_ai.data_cleaner.transformer_actions.constants import ImputationStrategy
{{ super() -}}
{% endblock %}
{% block action %}
    """
    Execute Transformer Action: ActionType.IMPUTE

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/actions/transformer_actions/README.md#fill-in-missing-values
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.IMPUTE,
        arguments=df.columns,  # Specify columns to impute
        axis=Axis.COLUMN,
        options={'strategy': ImputationStrategy.CONSTANT},  # Specify imputation strategy
    )
{% endblock %}



from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.constants import DATAFRAME_SAMPLE_COUNT_PREVIEW
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.data_preparation.utils.block.convert_content import convert_to_block
from mage_ai.server.api.base import BaseHandler
import asyncio
import json


class ApiPipelineBlockHandler(BaseHandler):
    def get(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        block = pipeline.get_block(block_uuid)
        include_outputs = self.get_bool_argument('include_outputs', True)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        self.write(
            dict(
                block=block.to_dict(
                    include_content=True,
                    include_outputs=include_outputs,
                )
            )
        )
        self.finish()

    def put(self, pipeline_uuid, block_uuid):
        """
        Allow updating block name, uuid, type, upstream_block, and downstream_blocks
        """
        pipeline = Pipeline.get(pipeline_uuid)
        data = json.loads(self.request.body).get('block', {})
        block = pipeline.get_block(block_uuid)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        block.update(data)
        self.write(dict(block=block.to_dict(include_content=True)))

    def delete(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        block = pipeline.get_block(block_uuid)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        block.delete()
        self.write(dict(block=block.to_dict()))


class ApiPipelineBlockExecuteHandler(BaseHandler):
    def post(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        block = pipeline.get_block(block_uuid)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        asyncio.run(block.execute())
        self.write(
            dict(
                block=block.to_dict(
                    include_outputs=True,
                )
            )
        )
        self.finish()


class ApiPipelineBlockListHandler(BaseHandler):
    def get(self, pipeline_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        include_outputs = self.get_bool_argument('include_outputs', True)
        self.write(
            dict(
                blocks=pipeline.to_dict(
                    include_content=True,
                    include_outputs=include_outputs,
                    sample_count=DATAFRAME_SAMPLE_COUNT_PREVIEW,
                )['blocks']
            )
        )
        self.finish()

    def post(self, pipeline_uuid):
        """
        Create block and add to pipeline
        """
        pipeline = Pipeline.get(pipeline_uuid)
        payload = json.loads(self.request.body).get('block', {})
        block = Block.create(
            payload.get('name') or payload.get('uuid'),
            payload.get('type'),
            get_repo_path(),
            config=payload.get('config'),
            configuration=payload.get('configuration'),
            language=payload.get('language'),
            pipeline=pipeline,
            priority=payload.get('priority'),
            upstream_block_uuids=payload.get('upstream_blocks', []),
        )

        content = payload.get('content')
        if content:
            if payload.get('converted_from'):
                content = convert_to_block(block, content)

            block.update_content(content)

        self.write(dict(block=block.to_dict(include_content=True)))


class ApiPipelineBlockAnalysisHandler(BaseHandler):
    def get(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        block = pipeline.get_block(block_uuid)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        analyses = block.get_analyses()
        self.write(dict(analyses=analyses))


class ApiPipelineBlockOutputHandler(BaseHandler):
    def get(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        block = pipeline.get_block(block_uuid)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        outputs = block.get_outputs(include_print_outputs=False, sample_count=None)
        self.write(dict(outputs=outputs))



if 'data_exporter' not in globals():
    from mage_ai.data_preparation.decorators import data_exporter


@data_exporter
def export_data_to_s3(df, **kwargs):
    """
    Template for exporting data to a S3 bucket.
    """
    s3_bucket = 'your_s3_bucket_name'
    s3_path_prefix = 'your_s3_path_prefix'
    (
        df.write
        .option('header', 'True')
        .mode('overwrite')
        # Use csv or parquet format
        .csv(f's3://{s3_bucket}/{s3_path_prefix}')
    )



"""
Useful conversion functions between Pandas DataFrames, numpy NDArrays, and ordered feature dictionaries
"""

from collections import OrderedDict
import logging
import numpy as np
import pandas as pd
import scipy
import six.moves

logger = logging.getLogger(__name__)


def df_to_fd(x, copy=False, nan_str='null', dtype=None):
    """
    Converts a pandas DataFrame to an ordered dictionary of numpy arrays

    :param x: The input DataFrame
    :param copy: Generates a copy of the underlying data
    :param nan_str: String used for nan values in DF
    :param dtype: The dtypes to cast to (if copying). Otherwise we don't cast.
        Can be a list (one per column) or a single dtype (applies to all features). Implies copy=True
    :return: feature dict representation of pandas DataFrame
    """
    assert isinstance(x, pd.DataFrame), 'input must be a Pandas DataFrame'

    # Find NaN values in object type columns
    nan_indices = {}
    for feat in x:
        if x[feat].dtype == np.dtype('O'):
            nan_indices[feat] = np.where(x[feat].isna().values)[0]

    if dtype and not copy:
        copy = True
    if copy:
        if dtype:
            if isinstance(dtype, list):
                assert len(dtype) == len(
                    list(x)
                ), 'dtypes must be same length as there are features in DataFrame'
                out = OrderedDict((f, x[f].values.astype(dtype[i])) for i, f in enumerate(list(x)))
                for feat in nan_indices:
                    out[feat][nan_indices[feat]] = nan_str
                return out
            else:
                out = OrderedDict((f, x[f].values.astype(dtype)) for f in list(x))
                for feat in nan_indices:
                    out[feat][nan_indices[feat]] = nan_str
                return out
        else:
            out = OrderedDict((f, x[f].values.copy()) for f in list(x))
            for feat in nan_indices:
                out[feat][nan_indices[feat]] = nan_str
            return out

    out = OrderedDict((f, x[f].values.view()) for f in list(x))
    for feat in nan_indices:
        try:
            out[feat][nan_indices[feat]] = nan_str
        except ValueError:
            logger.exception(feat, x[feat].dtype)
    return out


def np_to_fd(x, axes=None, feature_names=None, copy=True, dtype=None):
    """
    Converts a numpy ndarray to an ordered dictionary of numpy ndarrays by transposing
    and slicing on resulting rows (or any other axis)

    :param x: The ndarray. We will slice it across the transpose
    :param axes: How to reorder the axes of a tensor with rank > 2
        This takes a form of a tuple containing the desired ordering of resulting indices.
        We will slice features on the leading dimension following this transpose.

    Example:
    Converting an image batch of 100 1024Hx768W RGB (3 channels) image batch...
    The shape is (100,3,1024,768)

    If we want to extract 3 batches of 100 images each (one batch for each color) we'd do:
    axes=(1,0,2,3) since this maps to (channels=3, batch_size=100, height=1024, width=768)

    This will yield a transposed tensor of shape (3, 100, 1024, 768). We'll then slice on the leading dimension to yield a feature dict::

        {
            'red': ndarray with shape (100, 1024, 768)
            'blue': ndarray with shape (100, 1024, 768)
            'green': ndarray with shape (100, 1024, 768)
        }

    :param feature_names: A list of feature names to use. Otherwise we will auto-generate them
    :param copy: This makes a copy of the data rather than a view.
        We use row-major memory order so we get a contiguous array during column extraction
    :param dtype: The desired resulting numpy dtype (automatic if not specifying). Implies copy=True
    :return: FeatureDict
    """

    if dtype:
        copy = True
    assert isinstance(x, np.ndarray), 'input must be an ndarray'
    # pass-through 1D numpy array
    if x.ndim == 1:
        if feature_names:
            assert len(feature_names) == 1, '1D array detected but multiple feature names'
        fname = feature_names[0] if feature_names else 'c0'
        return OrderedDict([(fname, x)])
    view = x.view()
    view_t = np.transpose(view, axes=axes)
    # we usually want a copy since transpose
    # creates a non-contiguous array in memory for each feature we extract (which breaks some stuff)
    if copy:
        view_t = np.ascontiguousarray(view_t, dtype)
    cols = (
        feature_names
        if feature_names
        else ['c' + str(i) for i in six.moves.xrange(view_t.shape[0])]
    )
    return OrderedDict((cols[i], view_t[i]) for i in six.moves.xrange(view_t.shape[0]))


def sparse_to_dense(x):
    """
    Converts a sparse matrix into its dense representation. Passes through non-sparse matrices

    :param x: The input
    :return: A dense numpy array
    """
    if isinstance(x, scipy.sparse.csr_matrix):
        return x.todense()
    else:
        return x


def as_scalar(x, str_encoding='utf-8'):
    """
    Converts from a non-python scalar type to a Python scalar

    :param x: The scalar object to convert
    :param str_encoding: The encoding to use for converting fixed-width
    numpy strings (np.string_)
    """

    # types which need no further conversion
    PYTHON_SCALAR_TYPES = [
        float,
        int,
        bool,
        np.bool,  # an alias for Python bool but needed here separately
        str,
    ]
    NUMPY_SCALAR_TYPES = [np.bool_, np.str_, np.floating, np.integer]
    if type(x) in PYTHON_SCALAR_TYPES:
        return x
    elif isinstance(x, tuple(NUMPY_SCALAR_TYPES)):
        return x.item()
    elif isinstance(x, (np.string_, np.bytes_)):
        return x.item().decode(str_encoding)
    else:
        raise ValueError('Unknown type {}'.format(str(type(x))))


def fd_to_np(x, method='column_stack', axis=None, dtype=None):
    """
    Vertically stacks features together (by column)

    If there is a single feature, we return just that one

    This requires homogeneous shapes.

    :param x: The feature dict
    :param method: Either column_stack or stack. Refer to numpy docs for more info.
    :param axis: If using stack, the axis to stack on
    :param dtype: The dtype to cast to. Otherwise uses the numpy default
    :return: An ndarray containing the features in order
    """
    assert isinstance(x, OrderedDict), 'expected a feature dictionary'

    # passthrough if a feature
    if len(x) == 1:
        k = list(x.keys())[0]
        if dtype:
            return sparse_to_dense(x[k]).astype(dtype)
        return x[k]

    feature_names = list(x.keys())

    # shape of first feature
    # to ensure the leading dimensions are equal across features
    first_shape = None

    # check shapes and ensure everything is an ndarray
    for feature in feature_names:
        if isinstance(x[feature], np.ndarray):
            pass
        elif isinstance(x[feature], list):
            x[feature] = np.array(x[feature])
        elif isinstance(x[feature], scipy.sparse.csr_matrix):
            x[feature] = sparse_to_dense(x[feature])
        else:
            raise ValueError(
                'Unsupported container tensor type in feature dict: {}'.format(
                    str(type(x[feature]))
                )
            )
        if first_shape is None:
            first_shape = x[feature].shape
        assert first_shape[0] == x[feature].shape[0], 'incompatible shapes: {}, {}'.format(
            first_shape, x[feature].shape
        )
    if method == 'column_stack':
        out = np.column_stack([(x[key]) for key in x])
    elif method == 'stack':
        if axis:
            out = np.stack([(x[key]) for key in x], axis=axis)
        else:
            out = np.stack([(x[key]) for key in x], axis=-2)
    else:
        raise ValueError('unsupported method {}'.format(method))
    if dtype:
        return out.astype(dtype)
    return out


def fd_to_df(x):
    """
    :param x: FeatureDict
    :return: A Pandas DataFrame representation of the FeatureDict
    """
    assert isinstance(x, OrderedDict), 'expected a feature dictionary'
    return pd.DataFrame(x, columns=list(x.keys()))


def is_dtype(x, dtype):
    """
    Checks if a feature dict is all of a given type

    :param x: The feature dict
    :param dtype: The desired dtype
    :return: True if all elements are numpy arrays of the given dtype
    """
    for feat in x:
        if not isinstance(x[feat], np.ndarray):
            return False
        elif x[feat].dtype != dtype:
            return False
    return True


def cast_fd(x, dtype, order='C'):
    """
    Performs a cast of a dict containing numpy arrays

    :param x: The input ordereddict
    :param dtype: The target dtype
    :return: Casted FeatureDict
    """
    o = OrderedDict()
    for feat in x:
        if not isinstance(x[feat], np.ndarray):
            if isinstance(x[feat], list):
                o[feat] = np.array(x[feat]).astype(dtype=dtype, order=order)
            else:
                raise ValueError(
                    'Can only cast lists or numpy arrays. Key {} is of type {}'.format(
                        feat, str(type(x[feat]))
                    )
                )
        else:
            o[feat] = x[feat].astype(dtype=dtype, order=order)
    return o


# todo check memory layout C/Fortran
def to_fd(x, dtype=None):
    """
    Convenience function that auto-detects the type and performs the conversion

    :param x: Input data
    :param dtype: If None, we attempt to pass-through without copies. Doesn't apply to feature dicts
        Note: when passing non-numpy types do not set dtype.
    :return: A FeatureDict
    """
    if isinstance(x, OrderedDict):
        if dtype:
            if not is_dtype(x, dtype):
                return cast_fd(x, dtype)
        return x
    elif isinstance(x, np.ndarray):
        return np_to_fd(x, dtype=dtype)
    elif isinstance(x, pd.DataFrame):
        return df_to_fd(x, dtype=dtype)
    elif isinstance(x, dict):
        if dtype:
            if not is_dtype(x, dtype):
                return cast_fd(x, dtype)
        return OrderedDict((key, x[key]) for key in x)
    else:
        raise NotImplementedError('unknown type {}'.format(str(type(x))))


def to_list(feature):
    """
    Converts a feature of a vector type (pandas series or numpy ndarray)
    to a list native python types

    :param feature: The input feature as a list, numpy ndarray, or pandas Series
    :return: the corresponding data in native python type
    """
    if isinstance(feature, list):
        return [each.item() if isinstance(each, np.generic) else each for each in feature]
    elif isinstance(feature, (np.ndarray, pd.Series)):
        return feature.tolist()
    # handle the unlikely default case: feature_rows is not a list nor ndarray or pandas series
    else:
        return feature



from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.server.api.base import BaseHandler
from mage_ai.server.constants import VERSION


class ApiProjectsHandler(BaseHandler):
    def get(self):
        parts = get_repo_path().split('/')
        collection = [
            dict(
                name=parts[-1],
                version=VERSION,
            ),
        ]
        self.write(dict(projects=collection))



"""Add initial tables.

Revision ID: c07a23ff782b
Revises: 
Create Date: 2022-08-19 15:19:01.457470

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c07a23ff782b'
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('pipeline_schedule',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('pipeline_uuid', sa.String(length=255), nullable=True),
    sa.Column('schedule_type', sa.Enum('TIME', name='scheduletype'), nullable=True),
    sa.Column('start_time', sa.DateTime(timezone=True), nullable=True),
    sa.Column('schedule_interval', sa.String(length=50), nullable=True),
    sa.Column('status', sa.Enum('ACTIVE', 'INACTIVE', name='schedulestatus'), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('pipeline_run',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('pipeline_schedule_id', sa.Integer(), nullable=True),
    sa.Column('pipeline_uuid', sa.String(length=255), nullable=True),
    sa.Column('execution_date', sa.DateTime(timezone=True), nullable=True),
    sa.Column('status', sa.Enum('INITIAL', 'RUNNING', 'COMPLETED', 'FAILED', 'CANCELLED', name='pipelinerunstatus'), nullable=True),
    sa.ForeignKeyConstraint(['pipeline_schedule_id'], ['pipeline_schedule.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('block_run',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('pipeline_run_id', sa.Integer(), nullable=True),
    sa.Column('block_uuid', sa.String(length=255), nullable=True),
    sa.Column('status', sa.Enum('INITIAL', 'QUEUED', 'RUNNING', 'COMPLETED', 'FAILED', 'CANCELLED', name='blockrunstatus'), nullable=True),
    sa.ForeignKeyConstraint(['pipeline_run_id'], ['pipeline_run.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('block_run')
    op.drop_table('pipeline_run')
    op.drop_table('pipeline_schedule')
    # ### end Alembic commands ###



from mage_ai.data_cleaner.cleaning_rules.remove_outliers import RemoveOutliers
from mage_ai.tests.base_test import TestCase
import pandas as pd


class RemoveOutliersTests(TestCase):
    def test_evaluate(self):
        result = RemoveOutliers(**self.__create_test_data()).evaluate()

        self.assertEqual(
            result,
            [
                dict(
                    title='Remove outliers',
                    message='Remove 1 outlier(s) and null values to reduce the amount of noise in this column.',
                    status='not_applied',
                    action_payload=dict(
                        action_type='filter',
                        action_arguments=['number3'],
                        action_code='number3 <= 1779.01 and number3 >= 51.49000000000001',
                        action_options={},
                        action_variables={},
                        axis='row',
                        outputs=[],
                    ),
                ),
            ],
        )

    def test_evaluate_with_custom_config(self):
        test_data = self.__create_test_data()
        result = RemoveOutliers(
            test_data['df'],
            test_data['column_types'],
            test_data['statistics'],
            custom_config=dict(max_z_score=2),
        ).evaluate()

        self.assertEqual(
            result,
            [
                dict(
                    title='Remove outliers',
                    message='Remove 1 outlier(s) and null values to reduce the amount of noise in this column.',
                    status='not_applied',
                    action_payload=dict(
                        action_type='filter',
                        action_arguments=['number3'],
                        action_code='number3 <= 1491.0900000000001 and'
                        ' number3 >= 339.40999999999997',
                        action_options={},
                        action_variables={},
                        axis='row',
                        outputs=[],
                    ),
                ),
            ],
        )

    def test_evaluate_wrapped_column_name(self):
        df = pd.DataFrame(
            [
                ['1', '1', 1],
                ['1', '2', 997],
                ['1', '3', 998],
                ['1', '4', 999],
                ['1', '5', 1000],
                ['1', '6', 997],
                ['1', '7', 998],
                ['1', '8', 999],
                ['1', '9', 1000],
                ['1', '10', 997],
                ['1', '11', 998],
                ['1', '12', 999],
            ],
            columns=['number1.', 'num,ber2', 'number ([3])'],
        )
        column_types = {
            'number1.': 'text',
            'num,ber2': 'category',
            'number ([3])': 'number',
        }
        statistics = {
            'number ([3])/average': 915.25,
            'number ([3])/outlier_count': 1,
            'number ([3])/std': 287.92,
        }
        result = RemoveOutliers(df, column_types, statistics).evaluate()

        self.assertEqual(
            result,
            [
                dict(
                    title='Remove outliers',
                    message='Remove 1 outlier(s) and null values to reduce the amount of noise in this column.',
                    status='not_applied',
                    action_payload=dict(
                        action_type='filter',
                        action_arguments=['number ([3])'],
                        action_code='"number ([3])" <= 1779.01 and "number ([3])" >= 51.49000000000001',
                        action_options={},
                        action_variables={},
                        axis='row',
                        outputs=[],
                    ),
                )
            ],
        )

    def test_no_numerical_columns(self):
        df = pd.DataFrame(
            [
                ['1', '1', '1'],
                ['1', '2', '997'],
                ['1', '3', '998'],
                ['1', '4', '999'],
                ['1', '5', '1000'],
                ['1', '6', '997'],
                ['1', '7', '998'],
                ['1', '8', '999'],
                ['1', '9', '1000'],
                ['1', '10', '997'],
                ['1', '11', '998'],
                ['1', '12', '999'],
            ],
            columns=['number1.', 'num,ber2', 'number ([3])'],
        )
        column_types = {
            'number1.': 'text',
            'num,ber2': 'category',
            'number ([3])': 'zipcode',
        }
        statistics = {}
        result = RemoveOutliers(df, column_types, statistics).evaluate()
        self.assertEqual(result, [])

    def __create_test_data(self):
        df = pd.DataFrame(
            [
                [1, 1, 1],
                [1, 2, 997],
                [1, 3, 998],
                [1, 4, 999],
                [1, 5, 1000],
                [1, 6, 997],
                [1, 7, 998],
                [1, 8, 999],
                [1, 9, 1000],
                [1, 10, 997],
                [1, 11, 998],
                [1, 12, 999],
            ],
            columns=['number1', 'number2', 'number3'],
        )
        column_types = {
            'number1': 'number',
            'number2': 'number',
            'number3': 'number',
        }
        statistics = {
            'number1/average': 1,
            'number1/outlier_count': 0,
            'number1/std': 0,
            'number2/average': 6.5,
            'number2/outlier_count': 0,
            'number2/std': 3.6,
            'number3/average': 915.25,
            'number3/outlier_count': 1,
            'number3/std': 287.92,
        }
        return dict(
            df=df,
            column_types=column_types,
            statistics=statistics,
        )



from mage_ai.data_cleaner.shared.utils import is_spark_dataframe
from mage_ai.data_cleaner.transformer_actions import column, row
from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis, VariableType
from mage_ai.data_cleaner.transformer_actions.dependency_resolution import (
    default_resolution,
)
from mage_ai.data_cleaner.transformer_actions.helpers import drop_na
from mage_ai.data_cleaner.transformer_actions.variable_replacer import (
    interpolate,
    replace_true_false,
)
import json

try:
    from mage_ai.data_cleaner.transformer_actions.spark.transformers import (
        transform as transform_spark,
    )

    PYSPARK = True
except ImportError:
    PYSPARK = False

# from pipelines.column_type_pipelines import COLUMN_TYPE_PIPELINE_MAPPING

COLUMN_TYPE_PIPELINE_MAPPING = {}
DEPENDENCIES = {}
FUNCTION_MAPPING = {
    Axis.COLUMN: {
        ActionType.ADD: column.add_column,
        ActionType.AVERAGE: column.average,
        ActionType.CLEAN_COLUMN_NAME: column.clean_column_names,
        ActionType.COUNT: column.count,
        ActionType.COUNT_DISTINCT: column.count_distinct,
        ActionType.CUSTOM: column.custom,
        ActionType.DIFF: column.diff,
        # ActionType.EXPAND_COLUMN: column.expand_column,
        ActionType.FIRST: column.first,
        ActionType.FIX_SYNTAX_ERRORS: column.fix_syntax_errors,
        ActionType.IMPUTE: column.impute,
        ActionType.LAST: column.last,
        ActionType.MAX: column.max,
        ActionType.MEDIAN: column.median,
        ActionType.MIN: column.min,
        ActionType.REFORMAT: column.reformat,
        ActionType.REMOVE: column.remove_column,
        ActionType.REMOVE_OUTLIERS: column.remove_outliers,
        ActionType.SELECT: column.select,
        ActionType.SHIFT_DOWN: column.shift_down,
        ActionType.SHIFT_UP: column.shift_up,
        ActionType.SUM: column.sum,
    },
    Axis.ROW: {
        ActionType.CUSTOM: row.custom,
        ActionType.DROP_DUPLICATE: row.drop_duplicates,
        # ActionType.EXPLODE: row.explode,
        ActionType.FILTER: row.filter_rows,
        ActionType.SORT: row.sort_rows,
        ActionType.REMOVE: row.remove_row,
    },
}


class BaseAction:
    def __init__(self, action):
        self.action = action

        self.columns_by_type = {}
        for variable_data in self.action.get('action_variables', {}).values():
            if not variable_data:
                continue
            feature = variable_data.get(VariableType.FEATURE)
            if not feature:
                continue

            column_type = feature.get('column_type')
            if not self.columns_by_type.get(column_type):
                self.columns_by_type[column_type] = []
            self.columns_by_type[column_type].append(feature['uuid'])

    @property
    def action_type(self):
        return self.action['action_type']

    @property
    def axis(self):
        return self.action['axis']

    def execute(self, df, **kwargs):
        action_type = self.action['action_type']
        dependency = DEPENDENCIES.get(action_type, default_resolution)
        dependencies_met, msg = dependency(df, self.action)
        if not dependencies_met:
            raise RuntimeError(f'Dependencies of this cleaning action are not completed: {msg}')

        self.hydrate_action()

        if self.action.get('action_code'):
            self.action['action_code'] = replace_true_false(self.action['action_code'])

        if is_spark_dataframe(df):
            return self.execute_spark(df, **kwargs)

        if df.empty:
            return df

        if self.action_type in [ActionType.FILTER, ActionType.ADD]:
            df_transformed = self.transform(df)
        else:
            df_transformed = df

        if self.action_type == ActionType.GROUP:
            df_output = self.groupby(df, self.action)
        elif self.action_type == ActionType.JOIN:
            df_to_join = kwargs.get('df_to_join')
            df_output = self.join(df, df_to_join, self.action)
        else:
            column_types = {}
            for column_type, cols in self.columns_by_type.items():
                for col in cols:
                    column_types[col] = column_type
            df_output = FUNCTION_MAPPING[self.axis][self.action_type](
                df_transformed,
                self.action,
                column_types=column_types,
                original_df=df,
            )

        if self.action_type == ActionType.FILTER:
            return df.loc[df_output.index][df_output.columns]
        elif self.action_type == ActionType.ADD:
            output_cols = [f['uuid'] for f in self.action['outputs']]
            df[output_cols] = df_output[output_cols]
            return df
        else:
            return df_output

    def execute_spark(self, df, **kwargs):
        if not PYSPARK:
            raise RuntimeError('Spark is not supported in current environment')
        df = transform_spark(df, self.action, **kwargs).cache()
        return df

    def groupby(self, df, action):
        def __transform_partition(pdf, actions):
            for action in actions:
                pdf = BaseAction(action).execute(pdf)
            return pdf

        groupby_columns = action['action_arguments']
        return df.groupby(groupby_columns).apply(
            lambda x: __transform_partition(x, action['child_actions'])
        )

    def hydrate_action(self):
        for k, v in self.action['action_variables'].items():
            """
            k:
                1, 1_1
            v:
                {
                    'type': 'feature',
                    'id': 1,
                    'feature': {
                        'uuid': 'mage',
                    },
                }
            """
            if not v:
                continue
            if self.action.get('action_code'):
                self.action['action_code'] = interpolate(self.action['action_code'], k, v)

            if self.action.get('action_arguments'):
                self.action['action_arguments'] = [
                    interpolate(
                        args_text,
                        k,
                        v,
                    )
                    for args_text in self.action['action_arguments']
                ]

            if self.action.get('action_options'):
                action_options_json = json.dumps(self.action['action_options'])
                self.action['action_options'] = json.loads(interpolate(action_options_json, k, v))

    def join(self, df, df_to_join, action):
        action_options = action['action_options']
        left_on = action_options['left_on']
        right_on = action_options['right_on']

        for i in range(len(left_on)):
            col1, col2 = left_on[i], right_on[i]
            if df[col1].dtype != df_to_join[col2].dtype:
                df[col1] = drop_na(df[col1]).astype(str)
                df_to_join[col2] = drop_na(df_to_join[col2]).astype(str)

        if action.get('outputs') is not None:
            feature_rename_mapping = {
                f['source_feature']['uuid']: f['uuid']
                for f in action['outputs']
                if f.get('source_feature') is not None
            }
            df_to_join_renamed = df_to_join.rename(columns=feature_rename_mapping)
            right_on = [feature_rename_mapping.get(key, key) for key in right_on]
        else:
            df_to_join_renamed = df_to_join

        how = action_options.get('how', 'left')
        df_merged = df.merge(df_to_join_renamed, left_on=left_on, right_on=right_on, how=how)
        drop_columns = action_options.get('drop_columns', [])
        rename_columns = action_options.get('rename_columns', {})
        return df_merged.drop(columns=drop_columns).rename(columns=rename_columns)

    def transform(self, df):
        df_copy = df.copy()
        current_columns = df_copy.columns

        for column_type, original_columns in self.columns_by_type.items():
            cols = [col for col in original_columns if col in current_columns]

            if len(cols) == 0:
                continue

            build_pipeline = COLUMN_TYPE_PIPELINE_MAPPING.get(column_type)
            if not build_pipeline:
                continue

            df_copy[cols] = build_pipeline().fit_transform(df_copy[cols])

        return df_copy



"""Add variables to pipeline schedule.

Revision ID: 52ab80005742
Revises: c07a23ff782b
Create Date: 2022-08-25 10:31:48.480241

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '52ab80005742'
down_revision = 'c07a23ff782b'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('pipeline_schedule', sa.Column('variables', sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('pipeline_schedule', 'variables')
    # ### end Alembic commands ###



from enum import Enum
from mage_ai.data_cleaner.shared.utils import (
    is_geo_dataframe,
    is_spark_dataframe,
)
from mage_ai.data_preparation.models.constants import (
    DATAFRAME_ANALYSIS_KEYS,
    DATAFRAME_SAMPLE_COUNT,
    VARIABLE_DIR,
)
from mage_ai.shared.parsers import encode_complex
from typing import Any, Dict
import json
import simplejson
import os
import pandas as pd
import shutil


class VariableType(str, Enum):
    DATAFRAME = 'dataframe'
    DATAFRAME_ANALYSIS = 'dataframe_analysis'
    GEO_DATAFRAME = 'geo_dataframe'
    SPARK_DATAFRAME = 'spark_dataframe'


class Variable:
    def __init__(
        self,
        uuid: str,
        pipeline_path: str,
        block_uuid: str,
        partition: str = None,
        variable_type: VariableType = None
    ) -> None:
        self.uuid = uuid
        if not os.path.exists(pipeline_path):
            raise Exception(f'Pipeline {pipeline_path} does not exist.')
        self.pipeline_path = pipeline_path
        self.block_uuid = block_uuid
        self.partition = partition
        self.variable_dir_path = os.path.join(
            pipeline_path,
            VARIABLE_DIR,
            partition or '',
            block_uuid,
        )
        if not os.path.exists(self.variable_dir_path):
            os.makedirs(self.variable_dir_path)

        self.variable_type = variable_type
        self.check_variable_type()

    def check_variable_type(self):
        if self.variable_type is None and os.path.exists(
            os.path.join(self.variable_dir_path, f'{self.uuid}', 'data.parquet')
        ):
            # If parquet file exists for given variable, set the variable type to DATAFRAME
            self.variable_type = VariableType.DATAFRAME
        elif ((self.variable_type == VariableType.DATAFRAME or self.variable_type is None)
                and os.path.exists(
                os.path.join(self.variable_dir_path, f'{self.uuid}', 'data.sh'))):
            self.variable_type = VariableType.GEO_DATAFRAME

    @classmethod
    def dir_path(self, pipeline_path, block_uuid):
        return os.path.join(pipeline_path, VARIABLE_DIR, block_uuid)

    def delete(self):
        if self.variable_type is None and os.path.exists(
            os.path.join(self.variable_dir_path, f'{self.uuid}', 'data.parquet')
        ):
            # If parquet file exists for given variable, set the variable type to DATAFRAME
            self.variable_type = VariableType.DATAFRAME
        if self.variable_type == VariableType.DATAFRAME:
            self.__delete_parquet()
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            return self.__delete_dataframe_analysis()
        return self.__delete_json()

    def write_data(self, data: Any) -> None:
        if self.variable_type is None and type(data) is pd.DataFrame:
            self.variable_type = VariableType.DATAFRAME
        elif is_spark_dataframe(data):
            self.variable_type = VariableType.SPARK_DATAFRAME
        elif is_geo_dataframe(data):
            self.variable_type = VariableType.GEO_DATAFRAME

        if self.variable_type == VariableType.DATAFRAME:
            self.__write_parquet(data)
        elif self.variable_type == VariableType.SPARK_DATAFRAME:
            self.__write_spark_parquet(data)
        elif self.variable_type == VariableType.GEO_DATAFRAME:
            self.__write_geo_dataframe(data)
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            self.__write_dataframe_analysis(data)
        else:
            self.__write_json(data)

    def read_data(self, sample: bool = False, sample_count: int = None, spark=None) -> Any:
        self.check_variable_type()

        if self.variable_type == VariableType.DATAFRAME:
            return self.__read_parquet(sample=sample, sample_count=sample_count)
        elif self.variable_type == VariableType.SPARK_DATAFRAME:
            return self.__read_spark_parquet(sample=sample, sample_count=sample_count, spark=spark)
        elif self.variable_type == VariableType.GEO_DATAFRAME:
            return self.__read_geo_dataframe(sample=sample, sample_count=sample_count)
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            return self.__read_dataframe_analysis()
        return self.__read_json()

    def __delete_dataframe_analysis(self) -> None:
        variable_path = os.path.join(self.variable_dir_path, f'{self.uuid}')
        for k in DATAFRAME_ANALYSIS_KEYS:
            file_path = os.path.join(variable_path, f'{k}.json')
            if os.path.exists(file_path):
                os.remove(file_path)

    def __delete_json(self) -> None:
        file_path = os.path.join(self.variable_dir_path, f'{self.uuid}.json')
        if os.path.exists(file_path):
            os.remove(file_path)

    def __delete_parquet(self) -> None:
        variable_path = os.path.join(self.variable_dir_path, f'{self.uuid}')
        file_path = os.path.join(variable_path, 'data.parquet')
        if os.path.exists(file_path):
            os.remove(file_path)
            shutil.rmtree(variable_path)

    def __read_json(self, default_value={}) -> Dict:
        file_path = os.path.join(self.variable_dir_path, f'{self.uuid}.json')
        return self.__read_json_file(file_path, default_value)

    def __write_json(self, data) -> None:
        if not os.path.isdir(self.variable_dir_path):
            os.mkdir(self.variable_dir_path)
        self.__write_json_file(os.path.join(self.variable_dir_path, f'{self.uuid}.json'), data)

    def __read_json_file(self, file_path: str, default_value={}) -> Dict:
        if not os.path.exists(file_path):
            return default_value
        with open(file_path) as file:
            return json.load(file)

    def __write_json_file(self, file_path: str, data) -> None:
        with open(file_path, 'w') as file:
            simplejson.dump(
                data,
                file,
                default=encode_complex,
                ignore_nan=True,
            )

    def __read_geo_dataframe(self, sample: bool = False, sample_count: int = None):
        import geopandas as gpd

        variable_path = os.path.join(self.variable_dir_path, f'{self.uuid}')
        file_path = os.path.join(variable_path, 'data.sh')
        sample_file_path = os.path.join(variable_path, 'sample_data.sh')
        if not os.path.exists(file_path):
            return gpd.GeoDataFrame()
        if sample and os.path.exists(sample_file_path):
            try:
                df = gpd.read_file(sample_file_path)
            except Exception:
                df = gpd.read_file(file_path)
        else:
            df = gpd.read_file(file_path)
        if sample:
            sample_count = sample_count or DATAFRAME_SAMPLE_COUNT
            if df.shape[0] > sample_count:
                df = df.iloc[:sample_count]
        return df

    def __read_parquet(self, sample: bool = False, sample_count: int = None) -> pd.DataFrame:
        variable_path = os.path.join(self.variable_dir_path, f'{self.uuid}')
        file_path = os.path.join(variable_path, 'data.parquet')
        sample_file_path = os.path.join(variable_path, 'sample_data.parquet')
        if not os.path.exists(file_path):
            return pd.DataFrame()
        if sample and os.path.exists(sample_file_path):
            try:
                df = pd.read_parquet(sample_file_path, engine='pyarrow')
            except Exception:
                df = pd.read_parquet(file_path, engine='pyarrow')
        else:
            df = pd.read_parquet(file_path, engine='pyarrow')
        if sample:
            sample_count = sample_count or DATAFRAME_SAMPLE_COUNT
            if df.shape[0] > sample_count:
                df = df.iloc[:sample_count]
        return df

    def __read_spark_parquet(self, sample: bool = False, sample_count: int = None, spark=None):
        if spark is None:
            return None
        variable_path = os.path.join(self.variable_dir_path, f'{self.uuid}')
        return (
            spark.read
            .format('csv')
            .option('header', 'true')
            .option('inferSchema', 'true')
            .option('delimiter', ',')
            .load(variable_path)
        )

    def __write_geo_dataframe(self, data) -> None:
        variable_path = os.path.join(self.variable_dir_path, f'{self.uuid}')
        os.makedirs(variable_path, exist_ok=True)
        data.to_file(os.path.join(variable_path, 'data.sh'))
        df_sample_output = data.iloc[:DATAFRAME_SAMPLE_COUNT]
        df_sample_output.to_file(os.path.join(variable_path, 'sample_data.sh'))

    def __write_parquet(self, data: pd.DataFrame) -> None:
        df_output = data.copy()
        # Clean up data types since parquet doesn't support mixed data types
        for c in df_output.columns:
            series_non_null = df_output[c].dropna()
            if len(series_non_null) > 0:
                coltype = type(series_non_null.iloc[0])
                try:
                    df_output[c] = series_non_null.astype(coltype)
                except Exception:
                    # Fall back to convert to string
                    df_output[c] = series_non_null.astype(str)
        variable_path = os.path.join(self.variable_dir_path, f'{self.uuid}')
        os.makedirs(variable_path, exist_ok=True)
        df_output.to_parquet(os.path.join(variable_path, 'data.parquet'))
        df_sample_output = df_output.iloc[:DATAFRAME_SAMPLE_COUNT]
        df_sample_output.to_parquet(os.path.join(variable_path, 'sample_data.parquet'))

    def __write_spark_parquet(self, data) -> None:
        variable_path = os.path.join(self.variable_dir_path, f'{self.uuid}')
        (
            data.write
            .option('header', 'True')
            .mode('overwrite')
            .csv(variable_path)
        )

    def __read_dataframe_analysis(self) -> Dict[str, Dict]:
        """
        Read the following files
        1. metadata.json
        2. statistics.json
        3. insights.json
        4. suggestions.json
        """
        variable_path = os.path.join(self.variable_dir_path, f'{self.uuid}')
        if not os.path.exists(variable_path):
            return dict()
        result = dict()
        for k in DATAFRAME_ANALYSIS_KEYS:
            result[k] = self.__read_json_file(os.path.join(variable_path, f'{k}.json'))
        return result

    def __write_dataframe_analysis(self, data: Dict[str, Dict]) -> None:
        """
        Write the following files
        1. metadata.json
        2. statistics.json
        3. insights.json
        4. suggestions.json
        """
        variable_path = os.path.join(self.variable_dir_path, f'{self.uuid}')
        os.makedirs(variable_path, exist_ok=True)
        for k in DATAFRAME_ANALYSIS_KEYS:
            self.__write_json_file(os.path.join(variable_path, f'{k}.json'), data.get(k))



{% extends "transformers/transformer_actions/action.jinja" %}
{% block action %}
    """
    Execute Transformer Action: ActionType.FILTER

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/actions/transformer_actions/README.md#filter
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.FILTER,
        axis=Axis.ROW,
        action_code='',  # Specify your filtering code here
    )
{% endblock %}



import re


def camel_to_snake_case(name):
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    name = re.sub('__([A-Z])', r'_\1', name)
    name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', name)
    return name.lower()



from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.s3 import S3
from pandas import DataFrame
from os import path

if 'data_exporter' not in globals():
    from mage_ai.data_preparation.decorators import data_exporter


@data_exporter
def export_data_to_s3(df: DataFrame, **kwargs) -> None:
    """
    Template for exporting data to a S3 bucket.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/blocks/data_loading.md#s3
    """
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    bucket_name = 'your_bucket_name'
    object_key = 'your_object_key'

    S3.with_config(ConfigFileLoader(config_path, config_profile)).export(
        df,
        bucket_name,
        object_key,
    )
